@@ -2,6 +2,7 @@
 
 namespace onespace\tools\rabbit\behaviors;
 
+use Closure;
 use Exception;
 use InvalidArgumentException;
 use onespace\tools\rabbit\components\Rabbit;
@@ -23,9 +24,9 @@ use yii\helpers\Json;
  * @property Rabbit $rabbitComponent
  */
 class RabbitEventBehavior extends Behavior {
-    const EVENT_INSERT = 'insert';
-    const EVENT_UPDATE = 'update';
-    const EVENT_DELETE = 'delete';
+    public const EVENT_INSERT = 'insert';
+    public const EVENT_UPDATE = 'update';
+    public const EVENT_DELETE = 'delete';
 
     /**
      * Whether or not the behaviour should run
@@ -113,7 +114,7 @@ class RabbitEventBehavior extends Behavior {
     public $events = [
         self::EVENT_INSERT,
         self::EVENT_UPDATE,
-        self::EVENT_DELETE
+        self::EVENT_DELETE,
     ];
 
     /**
@@ -127,10 +128,25 @@ class RabbitEventBehavior extends Behavior {
     public $enforceCamelCase = false;
 
     /**
-     * Whether to use db transactions to maintain atomicity - ie, if sending to rabbit fails, undo the change locally
+     * Whether to use db transactions to maintain atomicity - ie, if sending to rabbit fails, undo the change locally.
+     * Is ignored if using outbox.
      */
     public $atomic = true;
 
+    /**
+     * Whether or not to use a rabbit outbox. Saves the rabbit query to a outbox table
+     * where a listener picks it up and sends it on.
+     */
+    public bool $useOutbox = true;
+
+    /**
+     * Any additional information to be passed to the rabbit query. Should
+     * be avoided if possible.
+     *
+     * If a `Closure` is passed, it should always return an array.
+     */
+    public Closure|array|null $extraData = null;
+
     /**
      * @var Transaction|null $transaction
      */
@@ -140,7 +156,6 @@ class RabbitEventBehavior extends Behavior {
      * {@inheritdoc}
      */
     public function events() {
-        Yii::info("EVENT RECEIVED", __METHOD__);
         if (!$this->liveSync) {
             return [];
         }
@@ -279,30 +294,138 @@ class RabbitEventBehavior extends Behavior {
         }
     }
 
+    /**
+     * This allows you to batch sync event payloads and send them in one transaction.
+     * 
+     * ## Important
+     * 
+     * 1. You must be sure your consumer can handle an array of the standard data format.
+     * 2. You intend this to hit a single queue on a single x-consistant-hash.
+     * 3. All the models passed in as a param must be of the same Class with the rabbit
+     *    behavior enabled. Not doing so will cause data to be emitted to an unintended target.
+     * 
+     * Use sparingly, you have been warned
+     * 
+     * @param \yii\base\Model[] $models Array of full models with the rabbit event behavior attached
+     * 
+     * ## Usage
+     * 
+     * ```php
+     * $models = MyModel::findAll([param => $param]);
+     * RabbitEventBehavior::bulkEmitSyncronisationEvent($models);
+     * ```
+     */
+    public static function bulkEmitSyncronisationEvent(array $models): void {
+        $payload = [];
+
+        foreach ($models as $model) {
+            try {
+                $payload[] = $model->prepareSyncMessage();
+            } catch (\Throwable $th) {
+                Yii::error($th->getMessage(), __METHOD__);
+                continue;
+            }
+        }
+
+        $lastModel = end($models);
+
+        $exception = null;
+        try {
+            if ($lastModel->useOutbox) {
+                // send the message to a defined exchange
+                $lastModel->rabbitComponent->outboxSendMessage(
+                    $lastModel->getExchangeName(),
+                    Json::encode($payload),
+                    $lastModel->getRabbitBehavourRoutingKey('sync'),
+                    ['hash-on' => $lastModel->getHashablePrimaryKey()]
+                );
+            } else {
+                $lastModel->rabbitComponent->sendMessage(
+                    $lastModel->rabbitComponent->declareExchange($lastModel->getExchangeName(), Rabbit::EXCHANGE_TYPE_TOPIC, [Rabbit::EXCHANGE_FLAG_DURABLE]),
+                    Json::encode($payload),
+                    $lastModel->getRabbitBehavourRoutingKey('sync'),
+                    ['hash-on' => $lastModel->getHashablePrimaryKey()]
+                );
+            }
+        } catch (Exception $ex) {
+            $exception = $ex;
+        }
+        // throw an exception to let the model know something went wrong
+        if ($exception !== null) {
+            throw new Exception('Event emmission error occurred: ' . $exception->getMessage(), 0, $exception->getCode(), $exception);
+        }
+    }
+
+    /**
+     * Common consistant way of creating the full payload sent on a emitSync
+     */
+    public function prepareSyncMessage(): array {
+        $message = [
+            'event' => [
+                'type' => 'sync',
+                'timestamp' => floor(microtime(true)),
+            ],
+            'entity' => [
+                'type' => $this->entityName,
+                'primaryKey' => $this->getPrimaryKey(),
+                'attributes' => $this->prepareAttributes(),
+                'relations' => $this->prepareRelations(),
+            ],
+        ];
+
+        if ($this->extraData !== null) {
+            $message['entity']['extra'] = $this->prepareExtra();
+        }
+
+        return $message;
+    }
+
     /**
      * Fire a manual sync event, which sends the current state of all attributes in the entity
      */
     public function emitSyncronisationEvent() {
         $exception = null;
         try {
-            $message = [
-                'event' => [
-                    'type' => 'sync',
-                    'timestamp' => floor(microtime(true)),
-                ],
-                'entity' => [
-                    'type' => $this->entityName,
-                    'primaryKey' => $this->getPrimaryKey(),
-                    'attributes' => $this->prepareAttributes(),
-                    'relations' => $this->prepareRelations(),
-                ],
-            ];
+            $message = $this->prepareSyncMessage();
+
+            if ($this->useOutbox) {
+                // send the message to a defined exchange
+                $this->rabbitComponent->outboxSendMessage(
+                    $this->getExchangeName(),
+                    Json::encode($message),
+                    $this->getRabbitBehavourRoutingKey('sync'),
+                    ['hash-on' => $this->getHashablePrimaryKey()]
+                );
+            } else {
+                $this->rabbitComponent->sendMessage(
+                    $this->rabbitComponent->declareExchange($this->getExchangeName(), Rabbit::EXCHANGE_TYPE_TOPIC, [Rabbit::EXCHANGE_FLAG_DURABLE]),
+                    Json::encode($message),
+                    $this->getRabbitBehavourRoutingKey('sync'),
+                    ['hash-on' => $this->getHashablePrimaryKey()]
+                );
+            }
+        } catch (Exception $ex) {
+            $exception = $ex;
+        }
+        // throw an exception to let the model know something went wrong
+        if ($exception !== null) {
+            throw new Exception('Event emmission error occurred: ' . $exception->getMessage(), 0, $exception->getCode(), $exception);
+        }
+    }
+
+    /**
+     * Similar to `emitSyncronisationEvent`, except the routing key is user defined. This allows a targeted sync event to a specified service
+     */
+    public function emitTargettedSyncronisationEvent(string $routingKey): void {
+        $exception = null;
+        try {
+            $message = $this->prepareSyncMessage();
 
             // send the message to a defined exchange
             $this->rabbitComponent->outboxSendMessage(
                 $this->getExchangeName(),
                 Json::encode($message),
-                $this->getRoutingKey('sync'),
+                $routingKey . implode('.', $this->getPrimaryKey()),
                 ['hash-on' => $this->getHashablePrimaryKey()]
             );
         } catch (Exception $ex) {
@@ -310,7 +433,7 @@ class RabbitEventBehavior extends Behavior {
         }
         // throw an exception to let the model know something went wrong
         if ($exception !== null) {
-            throw new Exception('Event emmission error occurred: ' . $exception->getMessage(), 0, $exception->getCode(), $exception);
+            throw new Exception('Event emmission error occurred: ' . $exception->getMessage(), $exception->getCode(), $exception);
         }
     }
 
@@ -360,7 +483,7 @@ class RabbitEventBehavior extends Behavior {
      */
     public function getHashablePrimaryKey() {
         $pk = $this->getPrimaryKey();
-        //sort the array by property name so that different orders don't throw things off
+        // sort the array by property name so that different orders don't throw things off
         ksort($pk);
         return implode('.', $pk);
     }
@@ -379,14 +502,18 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Prepare routing keys in a consistent way
      */
-    private function getRoutingKey(string $event) {
+    public function getRabbitBehavourRoutingKey(string $event) {
         return $this->entityName . ".$event." . implode('.', $this->getPrimaryKey());
     }
 
     /**
      * Prep a transaction to cover the event, so that we can undo it if for some reason the event isn't emitted
      */
-    function prepareTransaction() {
+    public function prepareTransaction() {
+        if ($this->useOutbox) {
+            // If using Rabbit Outbox, switch atomic off otherwise changes will never save.
+            $this->atomic = false;
+        }
         // if this is a standard yii db connection (not an azure one)
         if ($this->atomic && $this->owner->getDb() instanceof Connection) {
             $this->transaction = $this->owner->getDb()->beginTransaction();
@@ -398,7 +525,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Load up the attributes of the model, including any minor ones we want
      */
-    function prepareAttributes() {
+    public function prepareAttributes() {
         $return = [];
 
         $attributes = $this->owner->attributes;
@@ -410,7 +537,7 @@ class RabbitEventBehavior extends Behavior {
                     $attributes[$relationName] = [];
                     foreach ($relation->all() as $relatedModel) {
                         $pk = $this->getPrimaryKey();
-                        //sort the array by property name so that different orders don't throw things off
+                        // sort the array by property name so that different orders don't throw things off
                         ksort($pk);
                         $attributes[$relationName][implode('|', $pk)] = $relatedModel->attributes;
                     }
@@ -418,7 +545,7 @@ class RabbitEventBehavior extends Behavior {
                     // there is only a single relation. Merge them into a single object
                     // if there is no object there, create an empty one so the fields exist
                     if (($relatedModel = $relation->one()) == null) {
-                        $relatedModel = new $relation->modelClass;
+                        $relatedModel = new $relation->modelClass();
                     }
                     $attributes = ArrayHelper::merge($attributes, $relatedModel->attributes);
                 }
@@ -458,7 +585,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Load any relations that we want to pass
      */
-    function prepareRelations() {
+    protected function prepareRelations() {
         $return = [];
         foreach ($this->includeRelations as $relationName => $alias) {
             $relations = ArrayHelper::getValue($this->owner, $relationName);
@@ -498,16 +625,27 @@ class RabbitEventBehavior extends Behavior {
                     $return[$alias][] = $relationReturn;
                 }
             }
-            //$return[$alias] = [];
-
         }
         return $return;
     }
 
+    /**
+     * Get any extra data that should be passed to the queue.
+     */
+    protected function prepareExtra(): array {
+        if ($this->extraData instanceof Closure) {
+            $fn = $this->extraData;
+            $extra = $fn();
+        } else {
+            $extra = $this->extraData;
+        }
+        return $extra;
+    }
+
     /**
      * Emit a creation event to the server
      */
-    function afterInsert($event) {
+    public function afterInsert($event) {
         $exception = null;
         try {
             // get the current timestamp
@@ -526,14 +664,28 @@ class RabbitEventBehavior extends Behavior {
                 'author' => $this->getAuthor(),
             ];
 
+            if ($this->extraData !== null) {
+                $message['entity']['extra'] = $this->prepareExtra();
+            }
+
             // send the message to a defined exchange
-            $this->rabbitComponent->outboxSendMessage(
-                $this->getExchangeName(),
-                Json::encode($message),
-                $this->getRoutingKey('create'),
-                ['hash-on' => $this->getHashablePrimaryKey()],
-                $timestamp
-            );
+            if ($this->useOutbox) {
+                $this->rabbitComponent->outboxSendMessage(
+                    $this->getExchangeName(),
+                    Json::encode($message),
+                    $this->getRabbitBehavourRoutingKey('create'),
+                    ['hash-on' => $this->getHashablePrimaryKey()],
+                    $timestamp
+                );
+            } else {
+                $this->rabbitComponent->sendMessage(
+                    $this->rabbitComponent->declareExchange($this->getExchangeName(), Rabbit::EXCHANGE_TYPE_TOPIC, [Rabbit::EXCHANGE_FLAG_DURABLE]),
+                    Json::encode($message),
+                    $this->getRabbitBehavourRoutingKey('create'),
+                    ['hash-on' => $this->getHashablePrimaryKey()],
+                    $timestamp
+                );
+            }
             // job done
             if ($this->transaction != null) {
                 $this->transaction->commit();
@@ -555,7 +707,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Emit an update event to the rabbit server
      */
-    function afterUpdate(AfterSaveEvent $event) {
+    public function afterUpdate(AfterSaveEvent $event) {
         $exception = null;
         try {
             // get the new values for each changed attribute
@@ -582,7 +734,7 @@ class RabbitEventBehavior extends Behavior {
                 }
 
                 // check if the attribute is used in a constructed one
-                if (array_key_exists($rawKey, $this->constructedAttributes)) {
+                if (array_key_exists($rawKey, array_keys($this->constructedAttributes))) {
                     $constructed = $this->constructedAttributes[$rawKey];
                     if (is_array($constructed)) {
                         foreach ($constructed as $constructedAttr => $displayConstructedAttr) {
@@ -628,21 +780,28 @@ class RabbitEventBehavior extends Behavior {
                     'author' => $this->getAuthor(),
                 ];
 
-                Yii::info([
-                    'echange' => $this->getExchangeName(),
-                    'message' => Json::encode($message),
-                    'routingKey' => $this->getRoutingKey('update'),
-                    'hash' => ['hash-on' => $this->getHashablePrimaryKey()],
-                ], __METHOD__);
+                if ($this->extraData !== null) {
+                    $message['entity']['extra'] = $this->prepareExtra();
+                }
 
                 // send the message to a defined exchange
-                $this->rabbitComponent->outboxSendMessage(
-                    $this->getExchangeName(),
-                    Json::encode($message),
-                    $this->getRoutingKey('update'),
-                    ['hash-on' => $this->getHashablePrimaryKey()],
-                    $timestamp
-                );
+                if ($this->useOutbox) {
+                    $this->rabbitComponent->outboxSendMessage(
+                        $this->getExchangeName(),
+                        Json::encode($message),
+                        $this->getRabbitBehavourRoutingKey('update'),
+                        ['hash-on' => $this->getHashablePrimaryKey()],
+                        $timestamp
+                    );
+                } else {
+                    $this->rabbitComponent->sendMessage(
+                        $this->rabbitComponent->declareExchange($this->getExchangeName(), Rabbit::EXCHANGE_TYPE_TOPIC, [Rabbit::EXCHANGE_FLAG_DURABLE]),
+                        Json::encode($message),
+                        $this->getRabbitBehavourRoutingKey('update'),
+                        ['hash-on' => $this->getHashablePrimaryKey()],
+                        $timestamp
+                    );
+                }
             }
 
             // job done
@@ -666,7 +825,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Emit a deletion
      */
-    function afterDelete($event) {
+    public function afterDelete($event) {
         $exception = null;
         try {
             // get the current timestamp
@@ -682,18 +841,34 @@ class RabbitEventBehavior extends Behavior {
                 'entity' => [
                     'type' => $this->entityName,
                     'primaryKey' => $pk,
+                    'attributes' => $this->prepareAttributes(),
+                    'relations' => $this->prepareRelations(),
                 ],
                 'author' => $this->getAuthor(),
             ];
 
+            if ($this->extraData !== null) {
+                $message['entity']['extra'] = $this->prepareExtra();
+            }
+
             // send the message to a defined exchange
-            $this->rabbitComponent->outboxSendMessage(
-                $this->getExchangeName(),
-                Json::encode($message),
-                $this->getRoutingKey('delete'),
-                ['hash-on' => $this->getHashablePrimaryKey()],
-                $timestamp
-            );
+            if ($this->useOutbox) {
+                $this->rabbitComponent->outboxSendMessage(
+                    $this->getExchangeName(),
+                    Json::encode($message),
+                    $this->getRabbitBehavourRoutingKey('delete'),
+                    ['hash-on' => $this->getHashablePrimaryKey()],
+                    $timestamp
+                );
+            } else {
+                $this->rabbitComponent->sendMessage(
+                    $this->rabbitComponent->declareExchange($this->getExchangeName(), Rabbit::EXCHANGE_TYPE_TOPIC, [Rabbit::EXCHANGE_FLAG_DURABLE]),
+                    Json::encode($message),
+                    $this->getRabbitBehavourRoutingKey('delete'),
+                    ['hash-on' => $this->getHashablePrimaryKey()],
+                    $timestamp
+                );
+            }
             // job done
             if ($this->transaction != null) {
                 $this->transaction->commit();
@@ -715,7 +890,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Prep a transaction to cover the event, so that we can undo it if for some reason the event isn't emitted
      */
-    function prepareTransactionAsChild() {
+    public function prepareTransactionAsChild() {
         $this->parent = $this->childOf->one();
 
         foreach ($this->parent->behaviors as $behavior) {
@@ -731,7 +906,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Prep a transaction to cover the event, so that we can undo it if for some reason the event isn't emitted. Deletion as a child is a special case - we need to hold onto the current values of the attributes
      */
-    function prepareTransactionForDeletionAsChild() {
+    public function prepareTransactionForDeletionAsChild() {
         $this->attributeCache = $this->owner->attributes;
 
         return $this->prepareTransactionAsChild();
@@ -740,7 +915,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Emit a creation event to our parent model
      */
-    function afterInsertAsChild($event) {
+    public function afterInsertAsChild($event) {
         // an insertion counts as an update as far as the parent is concerned
         foreach ($this->parent->behaviors as $behavior) {
             if ($behavior instanceof self) {
@@ -761,7 +936,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Emit an update event to our parent model
      */
-    function afterUpdateAsChild(AfterSaveEvent $event) {
+    public function afterUpdateAsChild(AfterSaveEvent $event) {
         // let the parent know that certain attributes have changed
         foreach ($this->parent->behaviors as $behavior) {
             if ($behavior instanceof self) {
@@ -782,7 +957,7 @@ class RabbitEventBehavior extends Behavior {
     /**
      * Emit a deletion to our parent model
      */
-    function afterDeleteAsChild($event) {
+    public function afterDeleteAsChild($event) {
         // a deletion of a child would reflect as an update to the parent
         foreach ($this->parent->behaviors as $behavior) {
             if ($behavior instanceof self) {
@@ -801,5 +976,14 @@ class RabbitEventBehavior extends Behavior {
         }
     }
 
-    //hot reload test
+    /**
+     * To be used in `beforeSave` or `beforeDelete` in the case of a validation failure.
+     * If you are returning false, clear the transaction if it exists.
+     */
+    public function clearRabbitTransaction(): void {
+        if ($this->transaction !== null && $this->transaction->isActive) {
+            $this->transaction->rollBack();
+        }
+        $this->transaction = null;
+    }
 }
